/* disasm.c - disassemble 80836 code (CORE for dbug) */

#include <stdio.h>
#include <string.h>
#include "disasm.h"
#include "lowtype.h"

#define DEBUGGING
/* */

#ifndef NULL
# define NULL 0
#endif

#ifdef DEBUGGING
# define DEBUG_(func) fprintf func
#else
# define DEBUG_(func)
#endif

  /* naming convention:
   *   ops_XX         - decode instruction that may have operands
   *   prefix_XX      - decode instruction prefix of some kind
   *   ops_fix_XX     - one of the operands is fixed (given in table[Y].arg)
   *   ops_illegal    - illegal opcode
   *   ops_choose_d_w - add on d or w to opcode depending on operandsize
   *
   *      Letters (EGbv, etc) come from appendix A in "80386 Programmer's
   * Reference Manual" (1986).
   */

/* wanted output format:
 address  machine code
XXXXXXXX YYYYYYYYYYYYYYYYYYYYYY opcode  op1  ,op2  ,op3
 * where everything up to op1 is aligned.   The ops are allocated 5 positions,
 * but may be extended as necessary.  The address is left-justified.
 *  A printf that would output a line:
 * printf("%08x %-22x %-7s %-5s,%-5s,%s",address,machine,opcode,op1,op2,op3);
 * (Note: The above has bad var names, ignores max int sizes, and ignores
 *        the fact that we want to leave out unnecessary commas.)
 */

#define MAXSTRLEN 127

#define MAKE_CONST_ARG(arg,temp,func,addm,cntrlstr) do \
    { int cnst;                                       \
      void (*xyz_temporary)();                          \
      cnst=(*(func))();                               \
      (*(xyz_temporary=addm))(cnst);                  \
      sprintf((temp),(cntrlstr),cnst);                \
      (arg)=(temp);                                    \
    } while(0)

#define R_eAX 0
#define R_eCX 1
#define R_eDX 2
#define R_eBX 3
#define R_eSP 4
#define R_eBP 5
#define R_eSI 6
#define R_eDI 7

#define R_AL  0
#define R_CL  1
#define R_DL  2
#define R_BL  3
#define R_AH  4
#define R_CH  5
#define R_DH  6
#define R_BH  7

#define R_ES  0
#define R_CS  1
#define R_SS  2
#define R_DS  3
#define R_FS  4
#define R_GS  5

struct dis_instr
{ char instr[8];
  void (*func)();
  int arg;
};

typedef struct memory_info
{ int sz;
  unsigned char stf[8];
} memory_info_tp;

typedef int (*pointer_to_int_func)();

static void ops_2fix_AL_DX() ,ops_2fix_DX_AL(),ops_2fix_DX_eAX();
static void ops_2fix_eAX_DX(),ops_AL_Ib()     ,ops_AL_Ob()    ,ops_Ap();
static void ops_Cd_Rd()      ,ops_Dd_Rd()     ,ops_Eb()       ,ops_Eb_Gb();
static void ops_Eb_Ib()      ,ops_Ev()        ,ops_Ev_Gv()    ,ops_Ev_Gv_CL();
static void ops_Ev_Gv_Ib()   ,ops_Ev_Iv()     ,ops_Ew_Rw()    ,ops_Ew_Sw();
static void ops_Gb_Eb()      ,ops_Gv_Eb()     ,ops_Gv_Ev()    ,ops_Gv_Ev_Ib();
static void ops_Gv_Ev_Iv()   ,ops_Gv_Ew()     ,ops_Gv_M()     ,ops_Gv_Mp();
static void ops_Ib()         ,ops_Iv()        ,ops_Iw()       ,ops_Iw_Ib();
static void ops_Jb()         ,ops_Jv()        ,ops_Ob_AL()    ,ops_Ov_eAX();
static void ops_Rd_Cd()      ,ops_Rd_Dd()     ,ops_Rd_Td()    ,ops_Sw_Ew();
static void ops_Td_Rd()      ,ops_choose_d_w(),ops_eAX_Ib()   ,ops_eAX_Iv();
static void ops_eAX_Ov()     ,ops_escape()    ,ops_fix_breg_Ib();
static void ops_fix_eAX_vreg();
static void ops_fix_seg()    ,ops_fix_vreg()  ,ops_fix_vreg_Iv();
static void ops_float_d8()   ,ops_float_d9()  ,ops_float_da();
static void ops_float_db()   ,ops_float_dc()  ,ops_float_dd();
static void ops_float_de()   ,ops_float_df()  ,ops_grp_2_1();
static void ops_grp_2_CL()   ,ops_grp_3()     ,ops_grp_4()    ,ops_grp_5();
static void ops_grp_6()      ,ops_grp_7()     ,ops_grp_8()    ,ops_grp_Eb_Ib();
static void ops_grp_Ev_Ib()  ,ops_grp_Ev_Iv() ,ops_illegal()  ,ops_jcxz();
static void ops_none()       ,prefix_2_by()   ,prefix_asz()   ,prefix_lock();
static void prefix_osz()     ,prefix_rep()    ,prefix_repne() ,prefix_seg();

static struct dis_instr dis_tbl[256]
 ={ { "add"   ,ops_Eb_Gb             },  /* 0x00 */
    { "add"   ,ops_Ev_Gv             },  /* 0x01 */
    { "add"   ,ops_Gb_Eb             },  /* 0x02 */
    { "add"   ,ops_Gv_Ev             },  /* 0x03 */
    { "add"   ,ops_AL_Ib             },  /* 0x04 */
    { "add"   ,ops_eAX_Iv            },  /* 0x05 */
    { "push"  ,ops_fix_seg  ,R_ES    },  /* 0x06 */
    { "pop"   ,ops_fix_seg  ,R_ES    },  /* 0x07 */
    { "or"    ,ops_Eb_Gb             },  /* 0x08 */
    { "or"    ,ops_Ev_Gv             },  /* 0x09 */
    { "or"    ,ops_Gb_Eb             },  /* 0x0a */
    { "or"    ,ops_Gv_Ev             },  /* 0x0b */
    { "or"    ,ops_AL_Ib             },  /* 0x0c */
    { "or"    ,ops_eAX_Iv            },  /* 0x0d */
    { "push"  ,ops_fix_seg  ,R_CS    },  /* 0x0e */
    { ""      ,prefix_2_by           },  /* 0x0f */
    { "adc"   ,ops_Eb_Gb             },  /* 0x10 */
    { "adc"   ,ops_Ev_Gv             },  /* 0x11 */
    { "adc"   ,ops_Gb_Eb             },  /* 0x12 */
    { "adc"   ,ops_Gv_Ev             },  /* 0x13 */
    { "adc"   ,ops_AL_Ib             },  /* 0x14 */
    { "adc"   ,ops_eAX_Iv            },  /* 0x15 */
    { "push"  ,ops_fix_seg  ,R_SS    },  /* 0x16 */
    { "pop"   ,ops_fix_seg  ,R_SS    },  /* 0x17 */
    { "sbb"   ,ops_Eb_Gb             },  /* 0x18 */
    { "sbb"   ,ops_Ev_Gv             },  /* 0x19 */
    { "sbb"   ,ops_Gb_Eb             },  /* 0x1a */
    { "sbb"   ,ops_Gv_Ev             },  /* 0x1b */
    { "sbb"   ,ops_AL_Ib             },  /* 0x1c */
    { "sbb"   ,ops_eAX_Iv            },  /* 0x1d */
    { "push"  ,ops_fix_seg  ,R_DS    },  /* 0x1e */
    { "pop"   ,ops_fix_seg  ,R_DS    },  /* 0x1f */
    { "and"   ,ops_Eb_Gb             },  /* 0x20 */
    { "and"   ,ops_Ev_Gv             },  /* 0x21 */
    { "and"   ,ops_Gb_Eb             },  /* 0x22 */
    { "and"   ,ops_Gv_Ev             },  /* 0x23 */
    { "and"   ,ops_AL_Ib             },  /* 0x24 */
    { "and"   ,ops_eAX_Iv            },  /* 0x25 */
    { "es:"   ,prefix_seg   ,R_ES    },  /* 0x26 */
    { "daa"   ,ops_none              },  /* 0x27 */
    { "sub"   ,ops_Eb_Gb             },  /* 0x28 */
    { "sub"   ,ops_Ev_Gv             },  /* 0x29 */
    { "sub"   ,ops_Gb_Eb             },  /* 0x2a */
    { "sub"   ,ops_Gv_Ev             },  /* 0x2b */
    { "sub"   ,ops_AL_Ib             },  /* 0x2c */
    { "sub"   ,ops_eAX_Iv            },  /* 0x2d */
    { "cs:"   ,prefix_seg   ,R_CS    },  /* 0x2e */
    { "das"   ,ops_none              },  /* 0x2f */
    { "xor"   ,ops_Eb_Gb             },  /* 0x30 */
    { "xor"   ,ops_Ev_Gv             },  /* 0x31 */
    { "xor"   ,ops_Gb_Eb             },  /* 0x32 */
    { "xor"   ,ops_Gv_Ev             },  /* 0x33 */
    { "xor"   ,ops_AL_Ib             },  /* 0x34 */
    { "xor"   ,ops_eAX_Iv            },  /* 0x35 */
    { "ss:"   ,prefix_seg   ,R_SS    },  /* 0x36 */
    { "aaa"   ,ops_none              },  /* 0x37 */
    { "cmp"   ,ops_Eb_Gb             },  /* 0x38 */
    { "cmp"   ,ops_Ev_Gv             },  /* 0x39 */
    { "cmp"   ,ops_Gb_Eb             },  /* 0x3a */
    { "cmp"   ,ops_Gv_Ev             },  /* 0x3b */
    { "cmp"   ,ops_AL_Ib             },  /* 0x3c */
    { "cmp"   ,ops_eAX_Iv            },  /* 0x3d */
    { "ds:"   ,prefix_seg   ,R_DS    },  /* 0x3e */
    { "aas"   ,ops_none              },  /* 0x3f */
    { "inc"   ,ops_fix_vreg ,R_eAX   },  /* 0x40 */
    { "inc"   ,ops_fix_vreg ,R_eCX   },  /* 0x41 */
    { "inc"   ,ops_fix_vreg ,R_eDX   },  /* 0x42 */
    { "inc"   ,ops_fix_vreg ,R_eBX   },  /* 0x43 */
    { "inc"   ,ops_fix_vreg ,R_eSP   },  /* 0x44 */
    { "inc"   ,ops_fix_vreg ,R_eBP   },  /* 0x45 */
    { "inc"   ,ops_fix_vreg ,R_eSI   },  /* 0x46 */
    { "inc"   ,ops_fix_vreg ,R_eDI   },  /* 0x47 */
    { "dec"   ,ops_fix_vreg ,R_eAX   },  /* 0x48 */
    { "dec"   ,ops_fix_vreg ,R_eCX   },  /* 0x49 */
    { "dec"   ,ops_fix_vreg ,R_eDX   },  /* 0x4a */
    { "dec"   ,ops_fix_vreg ,R_eBX   },  /* 0x4b */
    { "dec"   ,ops_fix_vreg ,R_eSP   },  /* 0x4c */
    { "dec"   ,ops_fix_vreg ,R_eBP   },  /* 0x4d */
    { "dec"   ,ops_fix_vreg ,R_eSI   },  /* 0x4e */
    { "dec"   ,ops_fix_vreg ,R_eDI   },  /* 0x4f */
    { "push"  ,ops_fix_vreg ,R_eAX   },  /* 0x50 */
    { "push"  ,ops_fix_vreg ,R_eCX   },  /* 0x51 */
    { "push"  ,ops_fix_vreg ,R_eDX   },  /* 0x52 */
    { "push"  ,ops_fix_vreg ,R_eBX   },  /* 0x53 */
    { "push"  ,ops_fix_vreg ,R_eSP   },  /* 0x54 */
    { "push"  ,ops_fix_vreg ,R_eBP   },  /* 0x55 */
    { "push"  ,ops_fix_vreg ,R_eSI   },  /* 0x56 */
    { "push"  ,ops_fix_vreg ,R_eDI   },  /* 0x57 */
    { "pop"   ,ops_fix_vreg ,R_eAX   },  /* 0x58 */
    { "pop"   ,ops_fix_vreg ,R_eCX   },  /* 0x59 */
    { "pop"   ,ops_fix_vreg ,R_eDX   },  /* 0x5a */
    { "pop"   ,ops_fix_vreg ,R_eBX   },  /* 0x5b */
    { "pop"   ,ops_fix_vreg ,R_eSP   },  /* 0x5c */
    { "pop"   ,ops_fix_vreg ,R_eBP   },  /* 0x5d */
    { "pop"   ,ops_fix_vreg ,R_eSI   },  /* 0x5e */
    { "pop"   ,ops_fix_vreg ,R_eDI   },  /* 0x5f */
    { "pusha" ,ops_choose_d_w        },  /* 0x60 */
    { "popa"  ,ops_choose_d_w        },  /* 0x61 */
    { "bound" ,ops_Gv_Mp             },  /* 0x62 */
      /* bound is actually Gv_Ma, but my selected output form does Gv_Mp */
    { "arpl"  ,ops_Ew_Rw             },  /* 0x63 */
    { "fs:"   ,prefix_seg   ,R_FS    },  /* 0x64 */
    { "gs:"   ,prefix_seg   ,R_GS    },  /* 0x65 */
    { ""      ,prefix_osz            },  /* 0x66 */
    { ""      ,prefix_asz            },  /* 0x67 */
    { "push"  ,ops_Iv                },  /* 0x68 */
    { "imul"  ,ops_Gv_Ev_Iv          },  /* 0x69 */
    { "push"  ,ops_Ib                },  /* 0x6a */
    { "imul"  ,ops_Gv_Ev_Ib          },  /* 0x6b */
    { "insb"  ,ops_none              },  /* 0x6c */
    { "ins"   ,ops_choose_d_w        },  /* 0x6d */
    { "outsb" ,ops_none              },  /* 0x6e */
    { "outs"  ,ops_choose_d_w        },  /* 0x6f */
    { "jo"    ,ops_Jb                },  /* 0x70 */
    { "jno"   ,ops_Jb                },  /* 0x71 */
    { "jb"    ,ops_Jb                },  /* 0x72 */
    { "jnb"   ,ops_Jb                },  /* 0x73 */
    { "jz"    ,ops_Jb                },  /* 0x74 */
    { "jnz"   ,ops_Jb                },  /* 0x75 */
    { "jbe"   ,ops_Jb                },  /* 0x76 */
    { "ja"    ,ops_Jb                },  /* 0x77 */
    { "js"    ,ops_Jb                },  /* 0x78 */
    { "jns"   ,ops_Jb                },  /* 0x79 */
    { "jp"    ,ops_Jb                },  /* 0x7a */
    { "jnp"   ,ops_Jb                },  /* 0x7b */
    { "jl"    ,ops_Jb                },  /* 0x7c */
    { "jnl"   ,ops_Jb                },  /* 0x7d */
    { "jle"   ,ops_Jb                },  /* 0x7e */
    { "jg"    ,ops_Jb                },  /* 0x7f */
    { ""      ,ops_grp_Eb_Ib,1       },  /* 0x80 */
    { ""      ,ops_grp_Ev_Iv,1       },  /* 0x81 */
    { ""      ,ops_illegal           },  /* 0x82 */
    { ""      ,ops_grp_Ev_Ib,1       },  /* 0x83 */
    { "test"  ,ops_Eb_Gb             },  /* 0x84 */
    { "test"  ,ops_Ev_Gv             },  /* 0x85 */
    { "xchg"  ,ops_Eb_Gb             },  /* 0x86 */
    { "test"  ,ops_Ev_Gv             },  /* 0x87 */
    { "mov"   ,ops_Eb_Gb             },  /* 0x88 */
    { "mov"   ,ops_Ev_Gv             },  /* 0x89 */
    { "mov"   ,ops_Gb_Eb             },  /* 0x8a */
    { "mov"   ,ops_Gv_Ev             },  /* 0x8b */
    { "mov"   ,ops_Ew_Sw             },  /* 0x8c */
    { "lea"   ,ops_Gv_M              },  /* 0x8d */
    { "mov"   ,ops_Sw_Ew             },  /* 0x8e */
    { "pop"   ,ops_Ev                },  /* 0x8f */
    { "nop"   ,ops_none              },  /* 0x90 */
    { "xchg" ,ops_fix_eAX_vreg,R_eCX },  /* 0x91 */
    { "xchg" ,ops_fix_eAX_vreg,R_eDX },  /* 0x92 */
    { "xchg" ,ops_fix_eAX_vreg,R_eBX },  /* 0x93 */
    { "xchg" ,ops_fix_eAX_vreg,R_eSP },  /* 0x94 */
    { "xchg" ,ops_fix_eAX_vreg,R_eBP },  /* 0x95 */
    { "xchg" ,ops_fix_eAX_vreg,R_eSI },  /* 0x96 */
    { "xchg" ,ops_fix_eAX_vreg,R_eDI },  /* 0x97 */
    { "cbw"   ,ops_none              },  /* 0x98 */
    { "cwd"   ,ops_none              },  /* 0x99 */
    { "call"  ,ops_Ap                },  /* 0x9a */
    { "wait"  ,ops_none              },  /* 0x9b */
    { "pushf" ,ops_choose_d_w        },  /* 0x9c */
    { "popf"  ,ops_choose_d_w        },  /* 0x9d */
    { "sahf"  ,ops_none              },  /* 0x9e */
    { "lahf"  ,ops_none              },  /* 0x9f */
    { "mov"   ,ops_AL_Ob             },  /* 0xa0 */
    { "mov"   ,ops_eAX_Ov            },  /* 0xa1 */
    { "mov"   ,ops_Ob_AL             },  /* 0xa2 */
    { "mov"   ,ops_Ov_eAX            },  /* 0xa3 */
    { "movsb" ,ops_none              },  /* 0xa4 */
    { "movs"  ,ops_choose_d_w        },  /* 0xa5 */
    { "cmpsb" ,ops_none              },  /* 0xa6 */
    { "cmps"  ,ops_choose_d_w        },  /* 0xa7 */
    { "test"  ,ops_AL_Ib             },  /* 0xa8 */
    { "test"  ,ops_eAX_Iv            },  /* 0xa9 */
    { "stosb" ,ops_none              },  /* 0xaa */
    { "stos"  ,ops_choose_d_w        },  /* 0xab */
    { "lodsb" ,ops_none              },  /* 0xac */
    { "lods"  ,ops_choose_d_w        },  /* 0xad */
    { "scasb" ,ops_none              },  /* 0xae */
    { "scas"  ,ops_choose_d_w        },  /* 0xaf */
    { "mov"   ,ops_fix_breg_Ib,R_AL  },  /* 0xb0 */
    { "mov"   ,ops_fix_breg_Ib,R_CL  },  /* 0xb1 */
    { "mov"   ,ops_fix_breg_Ib,R_DL  },  /* 0xb2 */
    { "mov"   ,ops_fix_breg_Ib,R_BL  },  /* 0xb3 */
    { "mov"   ,ops_fix_breg_Ib,R_AH  },  /* 0xb4 */
    { "mov"   ,ops_fix_breg_Ib,R_CH  },  /* 0xb5 */
    { "mov"   ,ops_fix_breg_Ib,R_DH  },  /* 0xb6 */
    { "mov"   ,ops_fix_breg_Ib,R_BH  },  /* 0xb7 */
    { "mov"   ,ops_fix_vreg_Iv,R_eAX },  /* 0xb8 */
    { "mov"   ,ops_fix_vreg_Iv,R_eCX },  /* 0xb9 */
    { "mov"   ,ops_fix_vreg_Iv,R_eDX },  /* 0xba */
    { "mov"   ,ops_fix_vreg_Iv,R_eBX },  /* 0xbb */
    { "mov"   ,ops_fix_vreg_Iv,R_eSP },  /* 0xbc */
    { "mov"   ,ops_fix_vreg_Iv,R_eBP },  /* 0xbd */
    { "mov"   ,ops_fix_vreg_Iv,R_eSI },  /* 0xbe */
    { "mov"   ,ops_fix_vreg_Iv,R_eDI },  /* 0xbf */
    { ""      ,ops_grp_Eb_Ib  ,2     },  /* 0xc0 */
    { ""      ,ops_grp_Ev_Ib  ,2     },  /* 0xc1 */
    { "ret"   ,ops_Iw                },  /* 0xc2 */
    { "ret"   ,ops_none              },  /* 0xc3 */
    { "les"   ,ops_Gv_Mp             },  /* 0xc4 */
    { "lds"   ,ops_Gv_Mp             },  /* 0xc5 */
    { "mov"   ,ops_Eb_Ib             },  /* 0xc6 */
    { "mov"   ,ops_Ev_Iv             },  /* 0xc7 */
    { "enter" ,ops_Iw_Ib             },  /* 0xc8 */
    { "leave" ,ops_none              },  /* 0xc9 */
    { "retf"  ,ops_Iw                },  /* 0xca */
    { "retf"  ,ops_none              },  /* 0xcb */
    { "int3"  ,ops_none              },  /* 0xcc */
    { "int"   ,ops_Ib                },  /* 0xcd */
    { "into"  ,ops_none              },  /* 0xce */
    { "iret"  ,ops_none              },  /* 0xcf */
    { ""      ,ops_grp_2_1    ,1     },  /* 0xd0 */
    { ""      ,ops_grp_2_1    ,2     },  /* 0xd1 */
    { ""      ,ops_grp_2_CL   ,1     },  /* 0xd2 */
    { ""      ,ops_grp_2_CL   ,2     },  /* 0xd3 */
    { "aam"   ,ops_none              },  /* 0xd4 */
    { "aad"   ,ops_none              },  /* 0xd5 */
    { ""      ,ops_illegal           },  /* 0xd6 */
    { "xlat"  ,ops_none              },  /* 0xd7 */
    { ""      ,ops_float_d8          },  /* 0xd8 */
    { ""      ,ops_float_d9          },  /* 0xd9 */
    { ""      ,ops_float_da          },  /* 0xda */
    { ""      ,ops_float_db          },  /* 0xdb */
    { ""      ,ops_float_dc          },  /* 0xdc */
    { ""      ,ops_float_dd          },  /* 0xdd */
    { ""      ,ops_float_de          },  /* 0xde */
    { ""      ,ops_float_df          },  /* 0xdf */
    { "loopne",ops_Jb                },  /* 0xe0 */
    { "loope" ,ops_Jb                },  /* 0xe1 */
    { "loop"  ,ops_Jb                },  /* 0xe2 */
    { ""      ,ops_jcxz              },  /* 0xe3 */
    { "in"    ,ops_AL_Ib             },  /* 0xe4 */
    { "in"    ,ops_eAX_Ib            },  /* 0xe5 */
    { "out"   ,ops_AL_Ib             },  /* 0xe6 */
    { "out"   ,ops_eAX_Ib            },  /* 0xe7 */
    { "call"  ,ops_Jv                },  /* 0xe8 */
    { "jmp"   ,ops_Jv                },  /* 0xe9 */
    { "jmp"   ,ops_Ap                },  /* 0xea */
    { "jmp"   ,ops_Jb                },  /* 0xeb */
    { "in"    ,ops_2fix_AL_DX        },  /* 0xec */
    { "in"    ,ops_2fix_eAX_DX       },  /* 0xed */
    { "out"   ,ops_2fix_DX_AL        },  /* 0xee */
    { "out"   ,ops_2fix_DX_eAX       },  /* 0xef */
    { "lock"  ,prefix_lock           },  /* 0xf0 */
    { ""      ,ops_illegal           },  /* 0xf1 */
    { "repne" ,prefix_repne          },  /* 0xf2 */
    { "rep"   ,prefix_rep            },  /* 0xf3 */
    { "hlt"   ,ops_none              },  /* 0xf4 */
    { "cmc"   ,ops_none              },  /* 0xf5 */
    { ""      ,ops_grp_3      ,1     },  /* 0xf6 */
    { ""      ,ops_grp_3      ,4     },  /* 0xf7 */
    { "clc"   ,ops_none              },  /* 0xf8 */
    { "stc"   ,ops_none              },  /* 0xf9 */
    { "cli"   ,ops_none              },  /* 0xfa */
    { "sti"   ,ops_none              },  /* 0xfb */
    { "cld"   ,ops_none              },  /* 0xfc */
    { "std"   ,ops_none              },  /* 0xfd */
    { ""      ,ops_grp_4             },  /* 0xfe */
    { ""      ,ops_grp_5             }   /* 0xff */
  };

static struct dis_instr byte_2_disasm[256]
 ={ { ""      ,ops_grp_6             },  /* 2-0x00 */
    { ""      ,ops_grp_7             },  /* 2-0x01 */
    { "lar"   ,ops_Gv_Ew             },  /* 2-0x02 */
    { "lsl"   ,ops_Gv_Ew             },  /* 2-0x03 */
    { ""      ,NULL                  },  /* 2-0x04 */
    { ""      ,NULL                  },  /* 2-0x05 */
    { "clts"  ,ops_none              },  /* 2-0x06 */
    { ""      ,NULL                  },  /* 2-0x07 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x08-0x0b */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x0c-0x0f */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x10-0x13 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x14-0x17 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x18-0x1b */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x1c-0x1f */
    { "mov"   ,ops_Rd_Cd             },  /* 2-0x20 */
    { "mov"   ,ops_Rd_Dd             },  /* 2-0x21 */
    { "mov"   ,ops_Cd_Rd             },  /* 2-0x22 */
    { "mov"   ,ops_Dd_Rd             },  /* 2-0x23 */
    { "mov"   ,ops_Rd_Td             },  /* 2-0x24 */
    { ""      ,NULL                  },  /* 2-0x25 */
    { "mov"   ,ops_Td_Rd             },  /* 2-0x26 */
    { ""      ,NULL                  },  /* 2-0x27 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x28-0x2b */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x2c-0x2f */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x30-0x33 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x34-0x37 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x38-0x3b */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x3c-0x3f */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x40-0x43 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x44-0x47 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x48-0x4b */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x4c-0x4f */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x50-0x53 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x54-0x57 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x58-0x5b */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x5c-0x5f */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x60-0x63 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x64-0x67 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x68-0x6b */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x6c-0x6f */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x70-0x73 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x74-0x77 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x78-0x7b */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0x7c-0x7f */
    { "jo"    ,ops_Jv                },  /* 2-0x80 */
    { "jno"   ,ops_Jv                },  /* 2-0x81 */
    { "jb"    ,ops_Jv                },  /* 2-0x82 */
    { "jnb"   ,ops_Jv                },  /* 2-0x83 */
    { "jz"    ,ops_Jv                },  /* 2-0x84 */
    { "jnz"   ,ops_Jv                },  /* 2-0x85 */
    { "jbe"   ,ops_Jv                },  /* 2-0x86 */
    { "ja"    ,ops_Jv                },  /* 2-0x87 */
    { "js"    ,ops_Jv                },  /* 2-0x88 */
    { "jns"   ,ops_Jv                },  /* 2-0x89 */
    { "jp"    ,ops_Jv                },  /* 2-0x8a */
    { "jnp"   ,ops_Jv                },  /* 2-0x8b */
    { "jl"    ,ops_Jv                },  /* 2-0x8c */
    { "jnl"   ,ops_Jv                },  /* 2-0x8d */
    { "jle"   ,ops_Jv                },  /* 2-0x8e */
    { "jg"    ,ops_Jv                },  /* 2-0x8f */
    { "seto"  ,ops_Eb                },  /* 2-0x90 */
    { "setno" ,ops_Eb                },  /* 2-0x91 */
    { "setb"  ,ops_Eb                },  /* 2-0x92 */
    { "setnb" ,ops_Eb                },  /* 2-0x93 */
    { "setz"  ,ops_Eb                },  /* 2-0x94 */
    { "setnz" ,ops_Eb                },  /* 2-0x95 */
    { "setbe" ,ops_Eb                },  /* 2-0x96 */
    { "seta"  ,ops_Eb                },  /* 2-0x97 */
    { "sets"  ,ops_Eb                },  /* 2-0x98 */
    { "setns" ,ops_Eb                },  /* 2-0x99 */
    { "setp"  ,ops_Eb                },  /* 2-0x9a */
    { "setnp" ,ops_Eb                },  /* 2-0x9b */
    { "setl"  ,ops_Eb                },  /* 2-0x9c */
    { "setnl" ,ops_Eb                },  /* 2-0x9d */
    { "setle" ,ops_Eb                },  /* 2-0x9e */
    { "setg"  ,ops_Eb                },  /* 2-0x9f */
    { "push"  ,ops_fix_seg  ,R_FS    },  /* 2-0xa0 */
    { "pop"   ,ops_fix_seg  ,R_FS    },  /* 2-0xa1 */
    { ""      ,NULL                  },  /* 2-0xa2 */
    { "bt"    ,ops_Ev_Gv             },  /* 2-0xa3 */
    { "shld"  ,ops_Ev_Gv_Ib          },  /* 2-0xa4 */
    { "shld"  ,ops_Ev_Gv_CL          },  /* 2-0xa5 */
    { ""      ,NULL                  },  /* 2-0xa6 */
    { ""      ,NULL                  },  /* 2-0xa7 */
    { "push"  ,ops_fix_seg  ,R_GS    },  /* 2-0xa8 */
    { "pop"   ,ops_fix_seg  ,R_GS    },  /* 2-0xa9 */
    { ""      ,NULL                  },  /* 2-0xaa */
    { "bts"   ,ops_Ev_Gv             },  /* 2-0xab */
    { "shrd"  ,ops_Ev_Gv_Ib          },  /* 2-0xac */
    { "shrd"  ,ops_Ev_Gv_CL          },  /* 2-0xad */
    { ""      ,NULL                  },  /* 2-0xae */
    { "imul"  ,ops_Gv_Ev             },  /* 2-0xaf */
    { ""      ,NULL                  },  /* 2-0xb0 */
    { ""      ,NULL                  },  /* 2-0xb1 */
    { "lss"   ,ops_Gv_Mp             },  /* 2-0xb2 */
    { "btr"   ,ops_Ev_Gv             },  /* 2-0xb3 */
    { "lfs"   ,ops_Gv_Mp             },  /* 2-0xb4 */
    { "lgs"   ,ops_Gv_Mp             },  /* 2-0xb5 */
    { "movzx" ,ops_Gv_Eb             },  /* 2-0xb6 */
    { "movzx" ,ops_Gv_Ew             },  /* 2-0xb7 */
    { ""      ,NULL                  },  /* 2-0xb8 */
    { ""      ,NULL                  },  /* 2-0xb9 */
    { "lss"   ,ops_grp_8             },  /* 2-0xba */
    { "btc"   ,ops_Ev_Gv             },  /* 2-0xbb */
    { "bsf"   ,ops_Gv_Ev             },  /* 2-0xbc */
    { "bsr"   ,ops_Gv_Ev             },  /* 2-0xbd */
    { "movsx" ,ops_Gv_Eb             },  /* 2-0xbe */
    { "movsx" ,ops_Gv_Ew             },  /* 2-0xbf */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xc0-0xc3 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xc4-0xc7 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xc8-0xcb */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xcc-0xcf */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xd0-0xd3 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xd4-0xd7 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xd8-0xdb */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xdc-0xdf */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xe0-0xe3 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xe4-0xe7 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xe8-0xeb */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xec-0xef */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xf0-0xf3 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xf4-0xf7 */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xf8-0xfb */
    { "",NULL }, { "",NULL }, { "",NULL }, { "",NULL }, /* 2-0xfc-0xff */
  };

static struct dis_instr escape_tbl[8] /* alternate to replace floating point */
 ={ { "esc",ops_escape  }, /* 0xd8 */
    { "esc",ops_escape  }, /* 0xd9 */
    { "esc",ops_escape  }, /* 0xda */
    { "esc",ops_escape  }, /* 0xdb */
    { "esc",ops_escape  }, /* 0xdc */
    { "esc",ops_escape  }, /* 0xdd */
    { "esc",ops_escape  }, /* 0xde */
    { "esc",ops_escape  }  /* 0xdf */
  };

static struct dis_instr float_tbl[8]; /* initialized on 1st call */

static int last_arg; /* (see UGLY keyword below) */

static char out_address[MAXSTRLEN+1];
static char out_machine[MAXSTRLEN+1];
static int out_machine_len;
static char *out_opcode;
static char tempstr0[MAXSTRLEN+1]; /* used to compose opcode */
static char *out_arg1,*out_arg2,*out_arg3;
static int out_argcnt;
  /* tempstr[1-3] are used for composing out_arg[1-3] */
static char tempstr1[MAXSTRLEN+1],tempstr2[MAXSTRLEN+1],tempstr3[MAXSTRLEN+1];

static int default_sz,current_opsz,current_addsz;
static int status;

static pointer_to_int_func f_putaddress,f_stop,f_by,f_wo,f_dw,f_out,f_thisoff;
static char hex_convert[16]
    ={ '0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f' };
 /* NOTE: the C compiler balks if I try the intuitive "0123456789abcdef" by
  *   "too many initializers" or somesuch.  Can't it leave out the '\0'? */

static void addm_reset()
{ out_machine[0]='\0';
  out_machine_len=0;
}

static void addm_by(val)
{ char *ptr;

  if(out_machine_len>=MAXSTRLEN-1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"1H1\n"));
   return;
  }
  ptr=out_machine+out_machine_len;
  *ptr++=hex_convert[(val>>4)&0xf];
  *ptr++=hex_convert[val&0xf];
  *ptr='\0';
  out_machine_len+=2;
}

static void addm_wo(val)
{ char *ptr;

  if(out_machine_len>=MAXSTRLEN-2)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"2H1\n"));
   return;
  }
  ptr=out_machine+out_machine_len;
  *ptr++=hex_convert[(val>>4)&0xf];
  *ptr++=hex_convert[val&0xf];
  *ptr++=hex_convert[(val>>12)&0xf];
  *ptr++=hex_convert[(val>>8)&0xf];
  *ptr='\0';
  out_machine_len+=4;
}

static void addm_dw(val)
{ char *ptr;

  if(out_machine_len>=MAXSTRLEN-3)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"3H1\n"));
   return;
  }
  ptr=out_machine+out_machine_len;
  *ptr++=hex_convert[(val>>4)&0xf];
  *ptr++=hex_convert[val&0xf];
  *ptr++=hex_convert[(val>>12)&0xf];
  *ptr++=hex_convert[(val>>8)&0xf];
  *ptr++=hex_convert[(val>>20)&0xf];
  *ptr++=hex_convert[(val>>16)&0xf];
  *ptr++=hex_convert[(val>>28)&0xf];
  *ptr++=hex_convert[(val>>24)&0xf];
  *ptr='\0';
  out_machine_len+=8;
}

static void (*f_addm_dw)()=addm_dw,(*f_addm_wo)()=addm_wo;

static void do_output()
{ char out[7*MAXSTRLEN];
   /* in case all 6 parts are max, plus change for between */
 /* rough form:
  *  printf("%08x %-22x %-7s %-5s,%-5s,%s",address,machine,opcode,op1,op2,op3);
  */

  sprintf(out,"%8s %-22s %-7s",out_address,out_machine,out_opcode);
  switch(out_argcnt)
  { case 0:
    break;
    case 1:
     sprintf(out+strlen(out),"%-5s",out_arg1);
    break;
    case 2:
     sprintf(out+strlen(out),"%-5s,%-5s",out_arg1,out_arg2);
    break;
    case 3:
     sprintf(out+strlen(out),"%-5s,%-5s,%-5s",out_arg1,out_arg2,out_arg3);
    break;
    default:
     status=STATUS_INTERN_ERR;
     DEBUG_((stderr,"5H1\n"));
  }
  (*f_out)(out);
}

static void do_illegal_output()
{ out_opcode="db";
  out_argcnt=1;
  out_arg1=out_machine;
}

static int put_sreg(str,sreg) char str[MAXSTRLEN+1]; int sreg;
{ static char sregs[8][4]={ "es","cs","ss","ds","fs","gs","","" };

  if(sreg<0||sreg>=6) return 1;
  strcpy(str,sregs[sreg]);
 return 0;
}

static void put_reg(str,opsz,reg)
   char str[MAXSTRLEN+1];
   int opsz,reg;
{ static char regs[3][8][4]
     ={ { "al" ,"cl" ,"dl" ,"bl" ,"ah" ,"ch" ,"dh" ,"bh"  },
        { "ax" ,"cx" ,"dx" ,"bx" ,"sp" ,"bp" ,"si" ,"di"  },
        { "eax","ecx","edx","ebx","esp","ebp","esi","edi" }
      };

  if(reg<0||reg>=8)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"7H1\n"));
   return;
  }
  switch(opsz)
  { case 1:
     strcpy(str,regs[0][reg]);
    break;
    case 2:
     strcpy(str,regs[1][reg]);
    break;
    case 4:
     strcpy(str,regs[2][reg]);
    break;
    default:
     status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"8H1 %d\n",opsz));
    return;
  }
}

static void put_mem(str,opsz,modrm)
   char str[MAXSTRLEN+1];
   int opsz;
   memory_info_tp *modrm;
{ static char opsz_info[11][8]
    ={ ""     ,"by "  ,"wo ","UNK3 ","dw ","UNK5 ",
       "UNK6 ","UNK7 ","qw ","UNK9 ","tb "
     };

  if(modrm->sz<=0||modrm->sz>6||opsz>10||opsz<0)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"9H1\n"));
   return;
  }
  if((modrm->stf[0]&0xc0)==0xc0)       /* register */
  { if(modrm->sz!=1)
    { status=STATUS_INTERN_ERR;
      DEBUG_((stderr,"10H1\n"));
     return;
    }
    put_reg(str,opsz,modrm->stf[0]&0x7);
  }
  else if(current_addsz==2)           /* 16-bit addresses */
  { static char mod2by[8][8]
      ={ "bx+si","bx+di","bp+si","bp+di","si","di","bp","bx" };

    if((modrm->stf[0]&0xc7)==0x06)     /* 16-bit constant address */
    { sprintf(str,"%s[%x]",opsz_info[opsz],*(word *)(modrm->stf+1));
      if(modrm->sz!=3)
      { status=STATUS_INTERN_ERR;
        DEBUG_((stderr,"11H1\n"));
       return;
      }
    }
    else                             /* some regs. and possibly offset */
    { sprintf(str,"%s[%s",opsz_info[opsz],mod2by[modrm->stf[0]&0x7]);
      switch(modrm->stf[0]&0xc0)
      { case 0:
         sprintf(str+strlen(str),"]");
         if(modrm->sz!=1)
         { status=STATUS_INTERN_ERR;
           DEBUG_((stderr,"12H1\n"));
          return;
         }
        break;
        case 0x40:
         sprintf(str+strlen(str),"+%x]",(word)*(signed_byte *)(modrm->stf+1));
         if(modrm->sz!=2)
         { status=STATUS_INTERN_ERR;
           DEBUG_((stderr,"13H1\n"));
          return;
         }
        break;
        case 0x80:
         sprintf(str+strlen(str),"+%x]",*(word *)(modrm->stf+1));
         if(modrm->sz!=3)
         { status=STATUS_INTERN_ERR;
           DEBUG_((stderr,"14H1\n"));
          return;
         }
        break;
        default:
         status=STATUS_INTERN_ERR;
         DEBUG_((stderr,"15H1\n"));
        return;
      }
    }
  }
  else                               /* 32-bit addresses */
  { int constants_offset;

    if((modrm->stf[0]&0xc7)==0x05)     /* 32-bit constant address */
    { sprintf(str,"%s[%x]",opsz_info[opsz],*(dword *)(modrm->stf+1));
      if(modrm->sz!=5)
      { status=STATUS_INTERN_ERR;
        DEBUG_((stderr,"16H1\n"));
       return;
      }
     return;
    }
    else if((modrm->stf[0]&0x7)==4)    /* SIB byte present */
    { sprintf(str,"%s[",opsz_info[opsz]);
      if((modrm->stf[1]&0x38)!=0x20)   /* (scaled index present) */
      { static char scale_value[4][4]={ "","*2","*4","*8" };

        put_reg(str+strlen(str),4,(modrm->stf[1]>>3)&0x7);
        sprintf(str+strlen(str),"%s+",scale_value[(modrm->stf[1]>>6)&0x3]);
      }
      if((modrm->stf[1]&0x7)==5&&(modrm->stf[0]&0xc0)==0)
            /* special case with base!=ebp */
      { sprintf(str+strlen(str),"%x]",*(dword *)(modrm->stf+2));
        if(modrm->sz!=6)
        { status=STATUS_INTERN_ERR;
          DEBUG_((stderr,"16.5H1\n"));
        }
       return;
      }
      else
      { put_reg(str+strlen(str),4,modrm->stf[1]&0x7);
      }
      constants_offset=2;
    }
    else                               /* no SIB */
    { sprintf(str,"%s[",opsz_info[opsz]);
      put_reg(str+strlen(str),4,(modrm->stf[0])&0x7);
      constants_offset=1;
    }
    switch(modrm->stf[0]&0xc0)        /* final constant */
    { case 0:
       sprintf(str+strlen(str),"]");
       if(modrm->sz!=constants_offset)
       { status=STATUS_INTERN_ERR;
         DEBUG_((stderr,"17H1\n"));
        return;
       }
      break;
      case 0x40:
       sprintf(str+strlen(str),"+%x]"
          ,*(signed_byte *)(modrm->stf+constants_offset));
       if(modrm->sz!=constants_offset+1)
       { status=STATUS_INTERN_ERR;
         DEBUG_((stderr,"18H1\n"));
        return;
       }
      break;
      case 0x80:
       sprintf(str+strlen(str),"+%x]",*(dword *)(modrm->stf+constants_offset));
       if(modrm->sz!=constants_offset+4)
       { status=STATUS_INTERN_ERR;
         DEBUG_((stderr,"19H1\n"));
        return;
       }
      break;
      default:
       status=STATUS_INTERN_ERR;
       DEBUG_((stderr,"20H1 %d\n",modrm->stf[0]));
      return;
    }
  }
}

static void get_modrm(modrm)
   memory_info_tp *modrm;
{ modrm->stf[0]=(*f_by)();
  if(current_addsz==2)                      /* 16 bit */
  { if((modrm->stf[0]&0xc7)==6) goto is_word; /* constant 16-bit address */
    else
    { switch(modrm->stf[0]&0xc0)
      { case 0:
        case 0xc0:
         modrm->sz=1;
         addm_by(modrm->stf[0]);
        break;
        case 0x40:
         modrm->stf[1]=(*f_by)();
         addm_wo(*(word *)modrm->stf);
         modrm->sz=2;
        break;
        case 0x80:
        is_word:
         *(word *)(modrm->stf+1)=(*f_wo)();
         addm_wo(*(word *)modrm->stf);
         addm_by(modrm->stf[2]);
         modrm->sz=3;
        break;
        default:
         status=STATUS_INTERN_ERR;
         DEBUG_((stderr,"21H1\n"));
        return;
      }
    }
  }
  else                                       /* 32 bit */
  { if((modrm->stf[0]&0xc7)==5) goto is_dword; /* constant 32-bit address */
    else if((modrm->stf[0]&7)==4&&(modrm->stf[0]&0xc0)!=0xc0) /* SIB present */
    { modrm->stf[1]=(*f_by)();
      if((modrm->stf[1]&0x7)==5&&(modrm->stf[0]&0xc0)==0) goto is_scaleddword;
      switch(modrm->stf[0]&0xc0)
      { case 0:
        case 0xc0:
         addm_wo(*(word *)modrm->stf);
         modrm->sz=2;
        break;
        case 0x40:
         modrm->stf[2]=(*f_by)();
         addm_wo(*(word *)modrm->stf);
         addm_by(modrm->stf[2]);
         modrm->sz=3;
        break;
        case 0x80:
        is_scaleddword:
         *(dword *)(modrm->stf+2)=(*f_dw)();
         addm_dw(*(dword *)modrm->stf);
         addm_wo(*(word *)(modrm->stf+4));
         modrm->sz=6;
        break;
        default:
         status=STATUS_INTERN_ERR;
         DEBUG_((stderr,"22H1\n"));
        return;
      }
    }
    else
    { switch(modrm->stf[0]&0xc0)
      { case 0:
        case 0xc0:
         modrm->sz=1;
         addm_by(modrm->stf[0]);
        break;
        case 0x40:
         modrm->stf[1]=(*f_by)();
         addm_wo(*(word *)modrm->stf);
         modrm->sz=2;
        break;
        case 0x80:
        is_dword:
         *(dword *)(modrm->stf+1)=(*f_dw)();
         addm_dw(*(dword *)modrm->stf);
         addm_by(modrm->stf[4]);
         modrm->sz=5;
        break;
        default:
         status=STATUS_INTERN_ERR;
         DEBUG_((stderr,"23H1\n"));
        return;
      }
    }
  }
}

int dissassemble(fnc_out,fnc_putaddress,fnc_thisoff,fnc_stop,fnc_by,fnc_wo
                 ,fnc_dw,default_item_sz,allow_float)
   pointer_to_int_func fnc_out;
   pointer_to_int_func fnc_putaddress,fnc_thisoff; /* ret status */
   pointer_to_int_func fnc_stop;    /* ret 1 if no more dissassem. need */
   pointer_to_int_func fnc_by,fnc_wo,fnc_dw; /* ret by,wo,dw */
   int default_item_sz,allow_float;  /* 2 or 4 bytes */
{ static initialized;
  int op;

  if(!initialized)
  { memcpy(float_tbl,dis_tbl+0xd8,sizeof(float_tbl));
    initialized=1;
  }

  if(allow_float) memcpy(dis_tbl+0xd8,float_tbl,sizeof(float_tbl));
  else memcpy(dis_tbl+0xd8,escape_tbl,sizeof(float_tbl));

  f_out=fnc_out;
  f_putaddress=fnc_putaddress;  /* put this info where can be accessed */
  f_thisoff=fnc_thisoff;
  f_stop=fnc_stop;
  f_by=fnc_by; f_wo=fnc_wo; f_dw=fnc_dw;
  default_sz=default_item_sz;

  status=STATUS_OK;

  while(!(*f_stop)()&&status==STATUS_OK)
  { (*f_putaddress)(out_address,MAXSTRLEN); /* address */

    op=(*f_by)();  /* get opcode */

    addm_reset();   /* machine code */
    addm_by(op);

    out_opcode=dis_tbl[op].instr; /* not ALWAYS true, but overrideable */

      /* Store last arg; used only by ops_fix_seg().
       * UGLY way of doing this!!  Should probably track which
       * table op came from so that this isn't needed, and
       * never assume dis_tbl[] in functions.  However, for now
       * this hack should work. */
    last_arg=dis_tbl[op].arg;

    current_opsz=current_addsz=default_sz;
    (*(dis_tbl[op].func))(op);
    do_output();
  }
 return status;
}

static void process_another(table) struct dis_instr *table;
{ int op2;

  op2=(*f_by)();  /* get opcode */

  addm_by(op2);

  out_opcode=table[op2].instr;
  if(!table[op2].func)
  { do_illegal_output();
   return;
  }

    /* Store last arg; used only by ops_fix_seg().
     * UGLY way of doing this!!  Should probably track which
     * table op came from so that this isn't needed, and
     * never assume dis_tbl[] in functions.  However, for now
     * this hack should work. */
  last_arg=table[op2].arg;

  (*(table[op2].func))(op2);
}

   /* OPCODE-handling functions */

static void ops_2fix_AL_DX(op) int op;
{ out_argcnt=2;
  out_arg1="al";
  out_arg2="dx";
}

static void ops_2fix_DX_AL(op) int op;
{ out_argcnt=2;
  out_arg1="dx";
  out_arg2="al";
}

static void ops_2fix_DX_eAX(op) int op;
{ out_argcnt=2;
  out_arg1="dx";
  out_arg2=(current_opsz==4?"eax":"ax");
}

static void ops_2fix_eAX_DX(op) int op;
{ out_argcnt=2;
  out_arg1=(current_opsz==4?"eax":"ax");
  out_arg2="dx";
}

static void ops_AL_Ib(op) int op;
{ out_argcnt=2;
  out_arg1="al";
  MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
}

static void ops_AL_Ob(op) int op;
{ out_argcnt=2;
  out_arg1="al";
  MAKE_CONST_ARG(out_arg2,tempstr2,(current_addsz==4?f_dw:f_wo)
      ,(current_addsz==4?f_addm_dw:f_addm_wo),"by [%x]");
}

static void ops_Ap(op) int op;
{ int offset,segment;

  out_argcnt=1;

  if(current_opsz==4)
  { offset=(*f_dw)();
    addm_dw(offset);
  }
  else
  { offset=(*f_wo)();
    addm_wo(offset);
  }
  segment=(*f_wo)();
  addm_wo(segment);
  sprintf(tempstr1,"%x:%x",segment,offset);
  out_arg1=tempstr1;
}

static void ops_Cd_Rd(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"24H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { do_illegal_output();
   return;
  }

  out_argcnt=2;
  sprintf(tempstr1,"cr%d",((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,4,&modrm);
  out_arg2=tempstr2;
}

static void ops_Dd_Rd(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"25H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { do_illegal_output();
   return;
  }

  out_argcnt=2;
  sprintf(tempstr1,"dr%d",((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,4,&modrm);
  out_arg2=tempstr2;
}

static void ops_Eb(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"26H1\n"));
   return;
  }

  out_argcnt=1;
  put_mem(tempstr1,1,&modrm);
  out_arg1=tempstr1;
}

static void ops_Eb_Gb(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"27H1\n"));
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,1,&modrm);
  out_arg1=tempstr1;
  put_reg(tempstr2,1,((modrm.stf[0]>>3)&0x7));
  out_arg2=tempstr2;
}

static void ops_Eb_Ib(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"28H1\n"));
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,1,&modrm);
  out_arg1=tempstr1;
  MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
}

static void ops_Ev(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"29H1\n"));
   return;
  }

  out_argcnt=1;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
}

static void ops_Ev_Gv(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"30H1\n"));
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
  put_reg(tempstr2,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg2=tempstr2;
}

static void ops_Ev_Gv_CL(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"31H1\n"));
   return;
  }

  out_argcnt=3;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
  put_reg(tempstr2,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg2=tempstr2;
  out_arg3="cl";
}

static void ops_Ev_Gv_Ib(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"32H1\n"));
   return;
  }

  out_argcnt=3;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
  put_reg(tempstr2,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg2=tempstr2;
  MAKE_CONST_ARG(out_arg3,tempstr3,f_by,addm_by,"%x");
}

static void ops_Ev_Iv(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"33H1\n"));
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
  MAKE_CONST_ARG(out_arg2,tempstr2,(current_opsz==4?f_dw:f_wo)
      ,(current_opsz==4?f_addm_dw:f_addm_wo),"%x");
}

static void ops_Ew_Rw(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"35H1\n"));
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,2,&modrm);
  out_arg1=tempstr1;
  put_reg(tempstr2,2,((modrm.stf[0]>>3)&0x7));
  out_arg2=tempstr2;
}

static void ops_Ew_Sw(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"36H1\n"));
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,2,&modrm);
  out_arg1=tempstr1;
  if(put_sreg(tempstr2,((modrm.stf[0]>>3)&0x7)))
  { do_illegal_output();
   return;
  }
  out_arg2=tempstr2;
}

static void ops_Gb_Eb(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"37H1\n"));
   return;
  }

  out_argcnt=2;
  put_reg(tempstr1,1,((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,1,&modrm);
  out_arg2=tempstr2;
}

static void ops_Gv_Eb(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"38H1\n"));
   return;
  }

  out_argcnt=2;
  put_reg(tempstr1,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,1,&modrm);
  out_arg2=tempstr2;
}

static void ops_Gv_Ev(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"39H1\n"));
   return;
  }

  out_argcnt=2;
  put_reg(tempstr1,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,current_opsz,&modrm);
  out_arg2=tempstr2;
}

static void ops_Gv_Ev_Ib(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"40H1\n"));
   return;
  }

  out_argcnt=3;
  put_reg(tempstr1,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,current_opsz,&modrm);
  out_arg2=tempstr2;
  MAKE_CONST_ARG(out_arg3,tempstr3,f_by,addm_by,"%x");
}

static void ops_Gv_Ev_Iv(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"41H1\n"));
   return;
  }

  out_argcnt=3;
  put_reg(tempstr1,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,current_opsz,&modrm);
  out_arg2=tempstr2;
  MAKE_CONST_ARG(out_arg3,tempstr3,(current_opsz==4?f_dw:f_wo)
      ,(current_opsz==4?f_addm_dw:f_addm_wo),"%x");
}

static void ops_Gv_Ew(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"42H1\n"));
   return;
  }

  out_argcnt=2;
  put_reg(tempstr1,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,2,&modrm);
  out_arg2=tempstr2;
}

static void ops_Gv_M(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"43H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)==0xc0)
  { do_illegal_output();
   return;
  }

  out_argcnt=2;
  put_reg(tempstr1,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,0,&modrm);
  out_arg2=tempstr2;
}

static void ops_Gv_Mp(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"44H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)==0xc0)
  { do_illegal_output();
   return;
  }

  out_argcnt=2;
  put_reg(tempstr1,current_opsz,((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,current_opsz,&modrm);
  out_arg2=tempstr2;
}

static void ops_Ib(op) int op;
{ out_argcnt=1;
  MAKE_CONST_ARG(out_arg1,tempstr1,f_by,addm_by,"%x");
}

static void ops_Iv(op) int op;
{ out_argcnt=1;
  MAKE_CONST_ARG(out_arg1,tempstr1,(current_opsz==4?f_dw:f_wo)
      ,(current_opsz==4?f_addm_dw:f_addm_wo),"%x");
}

static void ops_Iw(op) int op;
{ out_argcnt=1;
  MAKE_CONST_ARG(out_arg1,tempstr1,f_wo,f_addm_wo,"%x");
}

static void ops_Iw_Ib(op) int op;
{ out_argcnt=2;
  MAKE_CONST_ARG(out_arg1,tempstr1,f_wo,f_addm_wo,"%x");
  MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
}

static void ops_Jb(op) int op;
{ int offset,posn;

  out_argcnt=1;

  offset=(signed_byte)(*f_by)();
  addm_by(offset);

  posn=offset+(*f_thisoff)();
  if(current_opsz==2)
  { posn&=0xffff;
  }
  sprintf(tempstr1,"%x (%x)",posn,offset);
  out_arg1=tempstr1;
}

static void ops_Jv(op) int op;
{ int offset,posn;
  out_argcnt=1;

  if(current_opsz==4)
  { offset=(*f_dw)();
    addm_dw(offset);

    posn=offset+(*f_thisoff)();
  }
  else
  { offset=(*f_wo)();
    addm_wo(offset);
    posn=(offset+(*f_thisoff)())&0xffff;
  }
  sprintf(tempstr1,"%x (%x)",posn,offset);
  out_arg1=tempstr1;
}

static void ops_Ob_AL(op) int op;
{ out_argcnt=2;
  MAKE_CONST_ARG(out_arg1,tempstr1,(current_addsz==4?f_dw:f_wo)
      ,(current_addsz==4?f_addm_dw:f_addm_wo),"by [%x]");
  out_arg2="al";
}

static void ops_Ov_eAX(op) int op;
{ out_argcnt=2;
  MAKE_CONST_ARG(out_arg1,tempstr1,(current_addsz==4?f_dw:f_wo)
    ,(current_addsz==4?f_addm_dw:f_addm_wo)
    ,(current_opsz==4?"dw [%x]":"wo [%x]"));
  out_arg2=(current_opsz==4?"eax":"ax");
}

static void ops_Rd_Cd(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"45H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { do_illegal_output();
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,4,&modrm);
  out_arg1=tempstr1;
  sprintf(tempstr2,"cr%d",((modrm.stf[0]>>3)&0x7));
  out_arg2=tempstr2;
}

static void ops_Rd_Dd(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"46H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { do_illegal_output();
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,4,&modrm);
  out_arg1=tempstr1;
  sprintf(tempstr2,"dr%d",((modrm.stf[0]>>3)&0x7));
  out_arg2=tempstr2;
}

static void ops_Rd_Td(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"47H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { do_illegal_output();
   return;
  }

  out_argcnt=2;
  put_mem(tempstr1,4,&modrm);
  out_arg1=tempstr1;
  sprintf(tempstr2,"tr%d",((modrm.stf[0]>>3)&0x7));
  out_arg2=tempstr2;
}

static void ops_Sw_Ew(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"36H1\n"));
   return;
  }

  out_argcnt=2;
  put_mem(tempstr2,2,&modrm);
  out_arg2=tempstr2;
  if(put_sreg(tempstr1,((modrm.stf[0]>>3)&0x7)))
  { do_illegal_output();
   return;
  }
  out_arg1=tempstr1;
}

static void ops_Td_Rd(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"48H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { do_illegal_output();
   return;
  }

  out_argcnt=2;
  sprintf(tempstr1,"tr%d",((modrm.stf[0]>>3)&0x7));
  out_arg1=tempstr1;
  put_mem(tempstr2,4,&modrm);
  out_arg2=tempstr2;
}

static void ops_choose_d_w(op) int op;
{ out_argcnt=0;
  sprintf(tempstr0,"%s%c",dis_tbl[op].instr,current_addsz==4?'d':'w');
  out_opcode=tempstr0;
}

static void ops_eAX_Ib(op) int op;
{ out_argcnt=2;
  out_arg1=current_opsz==4?"eax":"ax";
  MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
}

static void ops_eAX_Iv(op) int op;
{ out_argcnt=2;
  out_arg1=current_opsz==4?"eax":"ax";
  MAKE_CONST_ARG(out_arg2,tempstr2,(current_opsz==4?f_dw:f_wo)
      ,(current_opsz==4?f_addm_dw:f_addm_wo),"%x");
}

static void ops_eAX_Ov(op) int op;
{ out_argcnt=2;
  out_arg1=current_opsz==4?"eax":"ax";
  MAKE_CONST_ARG(out_arg2,tempstr2,(current_addsz==4?f_dw:f_wo)
    ,(current_addsz==4?f_addm_dw:f_addm_wo)
    ,(current_opsz==4?"dw [%x]":"wo [%x]"));
}

static void ops_escape(op) int op;
{ memory_info_tp modrm;
  int value;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"49H1\n"));
   return;
  }

  value=((op&0x7)<<3)+((modrm.stf[0]>>3)&0x7);

  out_argcnt=2;
  sprintf(tempstr1,"%x",value);
  out_arg1=tempstr1;
  put_mem(tempstr2,current_opsz,&modrm);
  out_arg2=tempstr2;
}

static void ops_fix_breg_Ib(op) int op;
{ out_argcnt=2;
  put_reg(tempstr1,1,dis_tbl[op].arg);
  out_arg1=tempstr1;
  MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
}

static void ops_fix_eAX_vreg(op) int op;
{ out_argcnt=2;
  out_arg1=(current_opsz==4?"eax":"ax");
  put_reg(tempstr2,current_opsz,dis_tbl[op].arg);
  out_arg2=tempstr2;
}

static void ops_fix_seg(op) int op;
{ out_argcnt=1;
  if(put_sreg(tempstr1,last_arg))
  { status=STATUS_INTERN_ERR;
   return;
  }
  out_arg1=tempstr1;
}

static void ops_fix_vreg(op) int op;
{ out_argcnt=1;
  put_reg(tempstr1,current_opsz,dis_tbl[op].arg);
  out_arg1=tempstr1;
}

static void ops_fix_vreg_Iv(op) int op;
{ out_argcnt=2;
  put_reg(tempstr1,current_opsz,dis_tbl[op].arg);
  out_arg1=tempstr1;
  MAKE_CONST_ARG(out_arg2,tempstr2,(current_opsz==4?f_dw:f_wo)
      ,(current_opsz==4?f_addm_dw:f_addm_wo),"%x");
}

static char float_arithname[8][8]
 ={ "add","mul","com","comp","sub","subr","div","divr" };

static void ops_float_d8(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"50H1\n"));
   return;
  }
  sprintf(tempstr0,"f%s",float_arithname[(modrm.stf[0]>>3)&0x7]);
  out_opcode=tempstr0;
  out_argcnt=2;
  out_arg1="st(0)";
  if((modrm.stf[0]&0xc0)==0xc0) sprintf(tempstr2,"st(%d)",(modrm.stf[0]&0x7));
  else put_mem(tempstr2,4,&modrm);
  out_arg2=tempstr2;
}

static void ops_float_d9(op) int op;
{ static char data1_d9[8][8]
    ={ "fld","","fst","fstp","fldenv","fldcw","fstenv","fstcw" };
  static char data2_d9[16][8]
    ={ "f2xm1"  ,"fyl2x"  ,"fptan","fpatan" ,"fxtract","fprem1",
       "fdecstp","fincstp","fprem","fyl2xp1","fsqrt"  ,"fsincos",
       "frndint","fscale" ,"fsin" ,"fcos"
     };
  static char data3_d9[16][8]
    ={ "fchs"  ,"fabs"  ,""    ,""      ,"ftst"  ,"fxam",
       ""      ,""      ,"fld1","fldl2t","fldl2e","fldpi",
       "fldlg2","fldln2","fldz",""
     };
  memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"51H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { if((modrm.stf[0]&0x38)==0x8)
    { do_illegal_output();
     return;
    }
    out_opcode=data1_d9[(modrm.stf[0]>>3)&0x7];
    out_argcnt=1;
    if((modrm.stf[0]&0x20)||current_opsz==2)
    { if(modrm.stf[0]&0x8) put_mem(tempstr1,2,&modrm);
      else put_mem(tempstr1,current_opsz,&modrm);
    }
    out_arg1=tempstr1;
  }
  else if((modrm.stf[0]&0xf0)==0xf0)
  { out_argcnt=0;
    out_opcode=data2_d9[modrm.stf[0]&0xf];
  }
  else if((modrm.stf[0]&0xf0)==0xc0)
  { if((modrm.stf[0]&0x8)==0x8)
    { out_opcode="fxch";
      out_argcnt=2;
      sprintf(tempstr1,"st(%d)",(modrm.stf[0]&0x7));
      out_arg1=tempstr1;
      out_arg2="st(0)";
    }
    else
    { out_opcode="fld";
      out_argcnt=1;
      sprintf(tempstr1,"st(%d)",(modrm.stf[0]&0x7));
      out_arg1=tempstr1;
    }
  }
  else if(modrm.stf[0]==0xd0)
  { out_opcode="fnop";
    out_argcnt=0;
  }
  else
  { out_argcnt=0;
    out_opcode=data3_d9[modrm.stf[0]&0xf];
    if(!*out_opcode)
    { do_illegal_output();
     return;
    }
  }
}

static void ops_float_da(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"52H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { sprintf(tempstr0,"fi%s",float_arithname[(modrm.stf[0]>>3)&0x7]);
    out_opcode=tempstr0;
    out_argcnt=2;
    out_arg1="st(0)";
    put_mem(tempstr2,4,&modrm);
    out_arg2=tempstr2;
  }
  else if(modrm.stf[0]==0xe9)
  { out_opcode="fucompp";
    out_argcnt=0;
  }
  else
  { do_illegal_output();
   return;
  }
}

static void ops_float_db(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"53H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { switch(modrm.stf[0]&0x38)
    { case 0:
       out_opcode="fild";
       out_argcnt=1;
       put_mem(tempstr1,4,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x10:
       out_opcode="fist";
       out_argcnt=1;
       put_mem(tempstr1,4,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x18:
       out_opcode="fistp";
       out_argcnt=1;
       put_mem(tempstr1,4,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x28:
       out_opcode="fld";
       out_argcnt=1;
       put_mem(tempstr1,10,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x38:
       out_opcode="fstp";
       out_argcnt=1;
       put_mem(tempstr1,10,&modrm);
       out_arg1=tempstr1;
      break;
      default:
       do_illegal_output();
      return;
    }
  }
  else
  { out_argcnt=0;
    switch(modrm.stf[0])
    { case 0xe0: out_opcode="feni"; break;
      case 0xe1: out_opcode="fdisi"; break;
      case 0xe2: out_opcode="fclex"; break;
      case 0xe3: out_opcode="finit"; break;
      case 0xe4: out_opcode="fsetpm"; break;
      default:
       do_illegal_output();
      return;
    }
  }
}

static void ops_float_dc(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"54H1\n"));
   return;
  }
  sprintf(tempstr0,"f%s",float_arithname[(modrm.stf[0]>>3)&0x7]);
  out_opcode=tempstr0;
  out_argcnt=2;
  if((modrm.stf[0]&0xc0)==0xc0)
  { if((modrm.stf[0]&0xf0)==0xd0)
    { do_illegal_output();
     return;
    }
    sprintf(tempstr1,"st(%d)",(modrm.stf[0]&0x7));
    out_arg1=tempstr1;
    out_arg2="st(0)";
  }
  else
  { out_arg1="st(0)";
    put_mem(tempstr2,4,&modrm);
    out_arg2=tempstr2;
  }
}

static void ops_float_dd(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"55H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { out_argcnt=1;
    switch(modrm.stf[0]&0x38)
    { case 0:
       out_opcode="fld";
       put_mem(tempstr1,8,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x10:
       out_opcode="fst";
       put_mem(tempstr1,8,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x18:
       out_opcode="fstp";
       put_mem(tempstr1,8,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x20:
       out_opcode="frstor";
       put_mem(tempstr1,current_opsz,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x30:
       out_opcode="fsave";
       put_mem(tempstr1,current_opsz,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x38:
       out_opcode="fstsw";
       put_mem(tempstr1,2,&modrm);
       out_arg1=tempstr1;
      break;
      default:
       do_illegal_output();
      return;
    }
  }
  else
  { static char data1_dd[8][8]
      ={ "ffree","","fst","fstp","fucom","fucomp","","" };

    out_argcnt=1;
    out_opcode=data1_dd[(modrm.stf[0]>>3)&0x7];
    if(!*out_opcode)
    { do_illegal_output();
     return;
    }
    sprintf(tempstr1,"st(%d)",(modrm.stf[0]&0x7));
    out_arg1=tempstr1;
  }
}

static void ops_float_de(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"56H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { sprintf(tempstr0,"fi%s",float_arithname[(modrm.stf[0]>>3)&0x7]);
    out_opcode=tempstr0;
    out_argcnt=2;
    out_arg1="st(0)";
    put_mem(tempstr2,2,&modrm);
    out_arg2=tempstr2;
  }
  else
  { switch(modrm.stf[0]&0x38)
    { case 0x18:
       if(modrm.stf[0]!=0xd9)
       { do_illegal_output();
        return;
       }
       /* PASS-THROUGH */
      case 0:
      case 0x8:
      case 0x20:
      case 0x28:
      case 0x30:
      case 0x38:
       sprintf(tempstr0,"f%sp",float_arithname[(modrm.stf[0]>>3)&0x7]);
       out_opcode=tempstr0;
       out_argcnt=2;
       put_mem(tempstr1,2,&modrm);
       out_arg1=tempstr1;
       out_arg2="st(0)";
      break;
      default:
       do_illegal_output();
      return;
    }
  }
}

static void ops_float_df(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"57H1\n"));
   return;
  }
  if((modrm.stf[0]&0xc0)!=0xc0)
  { out_argcnt=1;
    switch(modrm.stf[0]&0x38)
    { case 0:
       out_opcode="fild";
       put_mem(tempstr1,2,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x10:
       out_opcode="fist";
       put_mem(tempstr1,2,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x18:
       out_opcode="fistp";
       put_mem(tempstr1,2,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x20:
       out_opcode="fbld";
       put_mem(tempstr1,10,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x28:
       out_opcode="fild";
       put_mem(tempstr1,8,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x30:
       out_opcode="fbstp";
       put_mem(tempstr1,10,&modrm);
       out_arg1=tempstr1;
      break;
      case 0x38:
       out_opcode="fistp";
       put_mem(tempstr1,8,&modrm);
       out_arg1=tempstr1;
      break;
      default:
       do_illegal_output();
      return;
    }
  }
  else if(modrm.stf[0]==0xe0)
  { out_opcode="fstsw";
    out_argcnt=1;
    out_arg1="ax";
  }
  else
  { do_illegal_output();
   return;
  }
}

static char data_grps1_2[2][8][4]
  ={ { "add","or" ,"adc","sbb","and","sub","xor","cmp" },
     { "rol","ror","rcl","rcr","shl","shr",""   ,"sar" }
   };

static void ops_grp_2_1(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"58H1\n"));
   return;
  }

  out_opcode=data_grps1_2[2-1][(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=2;
  put_mem(tempstr1,dis_tbl[op].arg==1?1:current_opsz,&modrm);
  out_arg1=tempstr1;
  out_arg2="1";
}

static void ops_grp_2_CL(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"59H1\n"));
   return;
  }

  out_opcode=data_grps1_2[2-1][(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=2;
  put_mem(tempstr1,dis_tbl[op].arg==1?1:current_opsz,&modrm);
  out_arg1=tempstr1;
  out_arg2="cl";
}

static void ops_grp_3(op) int op;
{ memory_info_tp modrm;
  int opsz;

  if(dis_tbl[op].arg==1) opsz=1;
  else opsz=current_opsz;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"60H1\n"));
   return;
  }

  switch(modrm.stf[0]&0x38)
  { case 0:
     out_opcode="test";
     out_argcnt=2;
     put_mem(tempstr1,opsz,&modrm);
     out_arg1=tempstr1;
     if(dis_tbl[op].arg==1)
          MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
     else MAKE_CONST_ARG(out_arg2,tempstr2,(current_opsz==4?f_dw:f_wo)
            ,(current_opsz==4?f_addm_dw:f_addm_wo),"%x");
    break;
    case 0x10:
     out_opcode="not";
     out_argcnt=1;
     put_mem(tempstr1,opsz,&modrm);
     out_arg1=tempstr1;
    break;
    case 0x18:
     out_opcode="neg";
     out_argcnt=1;
     put_mem(tempstr1,opsz,&modrm);
     out_arg1=tempstr1;
    break;
    case 0x20:
     out_opcode="mul";
     out_argcnt=2;
     put_mem(tempstr1,opsz,&modrm);
     out_arg1=tempstr1;
     out_arg2=dis_tbl[op].arg==1?"al":(current_opsz==2?"ax":"eax");
    break;
    case 0x28:
     out_opcode="imul";
     out_argcnt=2;
     put_mem(tempstr1,opsz,&modrm);
     out_arg1=tempstr1;
     out_arg2=dis_tbl[op].arg==1?"al":(current_opsz==2?"ax":"eax");
    break;
    case 0x30:
     out_opcode="div";
     out_argcnt=2;
     put_mem(tempstr1,opsz,&modrm);
     out_arg1=tempstr1;
     out_arg2=dis_tbl[op].arg==1?"al":(current_opsz==2?"ax":"eax");
    break;
    case 0x38:
     out_opcode="idiv";
     out_argcnt=2;
     put_mem(tempstr1,opsz,&modrm);
     out_arg1=tempstr1;
     out_arg2=dis_tbl[op].arg==1?"al":(current_opsz==2?"ax":"eax");
    break;
    default:
     do_illegal_output();
    return;
  }
}

static void ops_grp_4(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"61H1\n"));
   return;
  }
  switch(modrm.stf[0]&0x38)
  { case 0:   out_opcode="inc"; break;
    case 0x8: out_opcode="dec"; break;
    default:
     do_illegal_output();
    return;
  }
  out_argcnt=1;
  put_mem(tempstr1,1,&modrm);
  out_arg1=tempstr1;
}

static void ops_grp_5(op) int op;
{ memory_info_tp modrm;
  static char *opcodes[8]
      ={ "inc","dec","call","call far","jmp","jmp far","push","" };

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"61H1\n"));
   return;
  }
  out_opcode=opcodes[(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=1;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
}

static void ops_grp_6(op) int op;
{ memory_info_tp modrm;
  static char data_grp6[8][8]
      ={ "sldt","str","lldt","ltr","verr","verw","","" };

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"62H1\n"));
   return;
  }

  out_opcode=data_grp6[(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=1;
  put_mem(tempstr1,2,&modrm);
  out_arg1=tempstr1;
}

static void ops_grp_7(op) int op;
{ memory_info_tp modrm;
  static char data_grp7[8][8]
      ={ "sgdt","sidt","lgdt","lidt","smsw","","lmsw","" };

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"63H1\n"));
   return;
  }

    /* if in register and not [sl]msw instr, its illegal! */
  if((modrm.stf[0]&0xc0)==0xc0&&(modrm.stf[0]&0xe0)!=0xe0)
  { do_illegal_output();
   return;
  }

  out_opcode=data_grp7[(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=1;
  put_mem(tempstr1,(modrm.stf[0]&0x38)>=0x20?2:0,&modrm);
  out_arg1=tempstr1;
}

static void ops_grp_8(op) int op;
{ memory_info_tp modrm;
  static char data_grp8[8][4]
      ={ "","","","","bt","bts","btr","btc" };

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"64H1\n"));
   return;
  }

  out_opcode=data_grp8[(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=2;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
  MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
}

static void ops_grp_Eb_Ib(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"65H1\n"));
   return;
  }

  out_opcode=data_grps1_2[dis_tbl[op].arg-1][(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=2;
  put_mem(tempstr1,1,&modrm);
  out_arg1=tempstr1;
  MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
}

static void ops_grp_Ev_Ib(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"66H1\n"));
   return;
  }

  out_opcode=data_grps1_2[dis_tbl[op].arg-1][(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=2;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
  MAKE_CONST_ARG(out_arg2,tempstr2,f_by,addm_by,"%x");
}

static void ops_grp_Ev_Iv(op) int op;
{ memory_info_tp modrm;

  get_modrm(&modrm);
  if(modrm.sz<1)
  { status=STATUS_INTERN_ERR;
    DEBUG_((stderr,"67H1\n"));
   return;
  }

  out_opcode=data_grps1_2[dis_tbl[op].arg-1][(modrm.stf[0]>>3)&0x7];
  if(!*out_opcode)
  { do_illegal_output();
   return;
  }
  out_argcnt=2;
  put_mem(tempstr1,current_opsz,&modrm);
  out_arg1=tempstr1;
  MAKE_CONST_ARG(out_arg2,tempstr2,(current_opsz==4?f_dw:f_wo)
      ,(current_opsz==4?f_addm_dw:f_addm_wo),"%x");
}

static void ops_illegal(op) int op;
{ do_illegal_output();
}

static void ops_jcxz(op) int op;
{ switch(current_opsz)
  { case 2:
     out_opcode="jcxz";
    break;
    case 4:
     out_opcode="jecxz";
    break;
    default:
     status=STATUS_INTERN_ERR;
     DEBUG_((stderr,"68H1\n"));
    return;
  }
  ops_Jb(op);
}

static void ops_none(op) int op;
{ out_argcnt=0;
}

static void prefix_2_by(op) int op;
{ process_another(byte_2_disasm);
}

static void prefix_asz(op) int op;
{ if(default_sz==2) current_addsz=4;
  else current_addsz=2;

  out_opcode="address size override:";
  out_argcnt=0;
  do_output();

  (*f_putaddress)(out_address,MAXSTRLEN); /* address */
  addm_reset();
  process_another(dis_tbl);
}

static void prefix_lock(op) int op;
{ out_argcnt=0;
}

static void prefix_osz(op) int op;
{ if(default_sz==2) current_opsz=4;
  else current_opsz=2;

  out_opcode="operand size override:";
  out_argcnt=0;
  do_output();

  (*f_putaddress)(out_address,MAXSTRLEN); /* address */
  addm_reset();
  process_another(dis_tbl);
}

static void prefix_rep(op) int op;
{ out_argcnt=0;
  do_output();

  (*f_putaddress)(out_address,MAXSTRLEN); /* address */
  addm_reset();
  process_another(dis_tbl);
}

static void prefix_repne(op) int op;
{ out_argcnt=0;
  do_output();

  (*f_putaddress)(out_address,MAXSTRLEN); /* address */
  addm_reset();
  process_another(dis_tbl);
}

static void prefix_seg(op) int op;
{ out_argcnt=0;
  do_output();

  (*f_putaddress)(out_address,MAXSTRLEN); /* address */
  addm_reset();
  process_another(dis_tbl);
}
